package failery

import (
	"bufio"
	"go/format"
	"io/ioutil"
	"path/filepath"
	"strings"
	"testing"

	"github.com/stretchr/testify/require"
	"github.com/stretchr/testify/suite"
)

const pkg = "test"

type GeneratorSuite struct {
	suite.Suite
	parser *Parser
}

func (s *GeneratorSuite) SetupTest() {
	s.parser = NewParser(nil)
}

func (s *GeneratorSuite) getInterfaceFromFile(interfacePath, interfaceName string) *Interface {
	if !strings.Contains(interfacePath, fixturePath) {
		interfacePath = filepath.Join(fixturePath, interfacePath)
	}
	s.NoError(
		s.parser.Parse(interfacePath), "The parser is able to parse the given file.",
	)

	s.NoError(
		s.parser.Load(), "The parser is able to load the config.",
	)

	iface, err := s.parser.Find(interfaceName)
	s.Require().NoError(err)
	s.Require().NotNil(iface)
	return iface
}

func (s *GeneratorSuite) getGenerator(
	filepath, interfaceName string, inPackage bool,
) *Generator {
	return NewGenerator(s.getInterfaceFromFile(filepath, interfaceName), pkg, inPackage)
}

func (s *GeneratorSuite) checkGeneration(
	filepath, interfaceName string, inPackage bool, expected string,
) *Generator {
	generator := s.getGenerator(filepath, interfaceName, inPackage)
	s.NoError(generator.Generate(), "The generator ran without errors.")

	// Mirror the formatting done by normally done by golang.org/x/tools/imports in Generator.Write.
	//
	// While we could possibly reuse Generator.Write here in addition to Generator.Generate,
	// it would require changing Write's signature to accept custom options, specifically to
	// allow the fragments in preexisting cases. It's assumed that this approximation,
	// just formatting the source, is sufficient for the needs of the current test styles.
	var actual []byte
	actual, fmtErr := format.Source(generator.buf.Bytes())
	s.NoError(fmtErr, "The formatter ran without errors.")

	// Compare lines for easier debugging via testify's slice diff output
	expectedLines := strings.Split(expected, "\n")
	actualLines := strings.Split(string(actual), "\n")

	s.Equal(
		expectedLines, actualLines,
		"The generator produced the expected output.",
	)
	return generator
}

func (s *GeneratorSuite) checkPrologueGeneration(
	generator *Generator, expected string,
) {
	generator.GeneratePrologue("mocks")
	s.Equal(
		expected, generator.buf.String(),
		"The generator produced the expected prologue.",
	)
}

func (s *GeneratorSuite) TestCalculateImport() {
	gp := []string{"a/src", "b/src"}

	s.Equal("c", calculateImport(gp, "a/src/c"))
	s.Equal("c", calculateImport(gp, "b/src/c"))
	s.Equal("d/src/c", calculateImport(gp, "d/src/c"))
}

func (s *GeneratorSuite) TestGenerator() {
	expected := `// Requester is an autogenerated failing mock type for the Requester type
type Requester struct {
	err error
}

// NewRequester creates a new Requester type instance
func NewRequester(err error) *Requester {
	return &Requester{err: err}
}

// Get provides a failing mock function with given fields: path
func (_m *Requester) Get(path string) (string, error) {
	var r0 string
	var r1 error
	r1 = _m.err

	return r0, r1
}
`
	s.checkGeneration(testFile, "Requester", false, expected)
}

func (s *GeneratorSuite) TestGeneratorSingleReturn() {
	expected := `// Requester2 is an autogenerated failing mock type for the Requester2 type
type Requester2 struct {
	err error
}

// NewRequester2 creates a new Requester2 type instance
func NewRequester2(err error) *Requester2 {
	return &Requester2{err: err}
}

// Get provides a failing mock function with given fields: path
func (_m *Requester2) Get(path string) error {
	var r0 error
	r0 = _m.err

	return r0
}
`
	s.checkGeneration(testFile2, "Requester2", false, expected)
}

func (s *GeneratorSuite) TestGeneratorNoArguments() {
	expected := `// Requester3 is an autogenerated failing mock type for the Requester3 type
type Requester3 struct {
	err error
}

// NewRequester3 creates a new Requester3 type instance
func NewRequester3(err error) *Requester3 {
	return &Requester3{err: err}
}

// Get provides a failing mock function with given fields:
func (_m *Requester3) Get() error {
	var r0 error
	r0 = _m.err

	return r0
}
`
	s.checkGeneration(
		filepath.Join(fixturePath, "requester3.go"), "Requester3", false,
		expected,
	)
}

func (s *GeneratorSuite) TestGeneratorNoNothing() {
	expected := `// Requester4 is an autogenerated failing mock type for the Requester4 type
type Requester4 struct {
	err error
}

// NewRequester4 creates a new Requester4 type instance
func NewRequester4(err error) *Requester4 {
	return &Requester4{err: err}
}

// Get provides a failing mock function with given fields:
func (_m *Requester4) Get() {
}
`
	s.checkGeneration(
		filepath.Join(fixturePath, "requester4.go"), "Requester4", false,
		expected,
	)
}

func (s *GeneratorSuite) TestGeneratorUnexported() {
	expected := `// failingRequester_unexported is an autogenerated failing mock type for the requester_unexported type
type failingRequester_unexported struct {
	err error
}

// NewFailingRequester_unexported creates a new failingRequester_unexported type instance
func NewFailingRequester_unexported(err error) *failingRequester_unexported {
	return &failingRequester_unexported{err: err}
}

// Get provides a failing mock function with given fields:
func (_m *failingRequester_unexported) Get() {
}
`
	s.checkGeneration(
		"requester_unexported.go", "requester_unexported", true, expected,
	)
}

func (s *GeneratorSuite) TestGeneratorPrologue() {
	generator := s.getGenerator(testFile, "Requester", false)
	expected := `package mocks

import test "github.com/kyma-project/kyma/tools/failery/failery/fixtures"

`
	s.checkPrologueGeneration(generator, expected)
}

func (s *GeneratorSuite) TestGeneratorPrologueWithImports() {
	generator := s.getGenerator("requester_ns.go", "RequesterNS", false)
	expected := `package mocks

import http "net/http"
import test "github.com/kyma-project/kyma/tools/failery/failery/fixtures"

`
	s.checkPrologueGeneration(generator, expected)
}

func (s *GeneratorSuite) TestGeneratorPrologueWithMultipleImportsSameName() {
	generator := s.getGenerator("same_name_imports.go", "Example", false)

	expected := `package mocks

import fixtureshttp "github.com/kyma-project/kyma/tools/failery/failery/fixtures/http"
import http "net/http"
import test "github.com/kyma-project/kyma/tools/failery/failery/fixtures"

`
	s.checkPrologueGeneration(generator, expected)
}

func (s *GeneratorSuite) TestGeneratorPrologueNote() {
	generator := s.getGenerator(testFile, "Requester", false)
	generator.GeneratePrologueNote("A\\nB")

	expected := `// Code generated by failery v1.0.0. DO NOT EDIT.

// A
// B

`

	s.Equal(expected, generator.buf.String())
}

func (s *GeneratorSuite) TestVersionOnCorrectLine() {
	gen := s.getGenerator(testFile, "Requester", false)

	//Run everything that is ran by the GeneratorVisitor
	gen.GeneratePrologueNote("A\\nB")
	gen.GeneratePrologue(pkg)
	err := gen.Generate()

	require.NoError(s.T(), err)
	scan := bufio.NewScanner(&gen.buf)
	s.Contains("Code generated by", scan.Text())
}

func (s *GeneratorSuite) TestGeneratorChecksInterfacesForNilable() {
	expected := `// RequesterIface is an autogenerated failing mock type for the RequesterIface type
type RequesterIface struct {
	err error
}

// NewRequesterIface creates a new RequesterIface type instance
func NewRequesterIface(err error) *RequesterIface {
	return &RequesterIface{err: err}
}

// Get provides a failing mock function with given fields:
func (_m *RequesterIface) Get() io.Reader {
	var r0 io.Reader

	return r0
}
`
	s.checkGeneration(
		filepath.Join(fixturePath, "requester_iface.go"), "RequesterIface",
		false, expected,
	)
}

func (s *GeneratorSuite) TestGeneratorPointers() {
	expected := `// RequesterPtr is an autogenerated failing mock type for the RequesterPtr type
type RequesterPtr struct {
	err error
}

// NewRequesterPtr creates a new RequesterPtr type instance
func NewRequesterPtr(err error) *RequesterPtr {
	return &RequesterPtr{err: err}
}

// Get provides a failing mock function with given fields: path
func (_m *RequesterPtr) Get(path string) (*string, error) {
	var r0 *string
	var r1 error
	r1 = _m.err

	return r0, r1
}
`
	s.checkGeneration(
		filepath.Join(fixturePath, "requester_ptr.go"), "RequesterPtr", false,
		expected,
	)
}

func (s *GeneratorSuite) TestGeneratorSlice() {
	expected := `// RequesterSlice is an autogenerated failing mock type for the RequesterSlice type
type RequesterSlice struct {
	err error
}

// NewRequesterSlice creates a new RequesterSlice type instance
func NewRequesterSlice(err error) *RequesterSlice {
	return &RequesterSlice{err: err}
}

// Get provides a failing mock function with given fields: path
func (_m *RequesterSlice) Get(path string) ([]string, error) {
	var r0 []string
	var r1 error
	r1 = _m.err

	return r0, r1
}
`
	s.checkGeneration(
		filepath.Join(fixturePath, "requester_slice.go"), "RequesterSlice",
		false, expected,
	)
}

func (s *GeneratorSuite) TestGeneratorArrayLiteralLen() {
	expected := `// RequesterArray is an autogenerated failing mock type for the RequesterArray type
type RequesterArray struct {
	err error
}

// NewRequesterArray creates a new RequesterArray type instance
func NewRequesterArray(err error) *RequesterArray {
	return &RequesterArray{err: err}
}

// Get provides a failing mock function with given fields: path
func (_m *RequesterArray) Get(path string) ([2]string, error) {
	var r0 [2]string
	var r1 error
	r1 = _m.err

	return r0, r1
}
`
	s.checkGeneration(
		filepath.Join(fixturePath, "requester_array.go"), "RequesterArray",
		false, expected,
	)
}

func (s *GeneratorSuite) TestGeneratorNamespacedTypes() {
	expected := `// RequesterNS is an autogenerated failing mock type for the RequesterNS type
type RequesterNS struct {
	err error
}

// NewRequesterNS creates a new RequesterNS type instance
func NewRequesterNS(err error) *RequesterNS {
	return &RequesterNS{err: err}
}

// Get provides a failing mock function with given fields: path
func (_m *RequesterNS) Get(path string) (http.Response, error) {
	var r0 http.Response
	var r1 error
	r1 = _m.err

	return r0, r1
}
`
	s.checkGeneration(
		filepath.Join(fixturePath, "requester_ns.go"), "RequesterNS", false,
		expected,
	)
}

func (s *GeneratorSuite) TestGeneratorWhereArgumentNameConflictsWithImport() {
	expected := `// RequesterArgSameAsImport is an autogenerated failing mock type for the RequesterArgSameAsImport type
type RequesterArgSameAsImport struct {
	err error
}

// NewRequesterArgSameAsImport creates a new RequesterArgSameAsImport type instance
func NewRequesterArgSameAsImport(err error) *RequesterArgSameAsImport {
	return &RequesterArgSameAsImport{err: err}
}

// Get provides a failing mock function with given fields: _a0
func (_m *RequesterArgSameAsImport) Get(_a0 string) *json.RawMessage {
	var r0 *json.RawMessage

	return r0
}
`
	s.checkGeneration(
		filepath.Join(fixturePath, "requester_arg_same_as_import.go"),
		"RequesterArgSameAsImport", false, expected,
	)
}

func (s *GeneratorSuite) TestGeneratorWhereArgumentNameConflictsWithNamedImport() {
	expected := `// RequesterArgSameAsNamedImport is an autogenerated failing mock type for the RequesterArgSameAsNamedImport type
type RequesterArgSameAsNamedImport struct {
	err error
}

// NewRequesterArgSameAsNamedImport creates a new RequesterArgSameAsNamedImport type instance
func NewRequesterArgSameAsNamedImport(err error) *RequesterArgSameAsNamedImport {
	return &RequesterArgSameAsNamedImport{err: err}
}

// Get provides a failing mock function with given fields: _a0
func (_m *RequesterArgSameAsNamedImport) Get(_a0 string) *json.RawMessage {
	var r0 *json.RawMessage

	return r0
}
`
	s.checkGeneration(
		filepath.Join(fixturePath, "requester_arg_same_as_named_import.go"),
		"RequesterArgSameAsNamedImport", false, expected,
	)
}

func (s *GeneratorSuite) TestGeneratorWhereArgumentNameConflictsWithPackage() {
	expected := `// RequesterArgSameAsPkg is an autogenerated failing mock type for the RequesterArgSameAsPkg type
type RequesterArgSameAsPkg struct {
	err error
}

// NewRequesterArgSameAsPkg creates a new RequesterArgSameAsPkg type instance
func NewRequesterArgSameAsPkg(err error) *RequesterArgSameAsPkg {
	return &RequesterArgSameAsPkg{err: err}
}

// Get provides a failing mock function with given fields: _a0
func (_m *RequesterArgSameAsPkg) Get(_a0 string) {
}
`
	s.checkGeneration(
		filepath.Join(fixturePath, "requester_arg_same_as_pkg.go"),
		"RequesterArgSameAsPkg", false, expected,
	)
}

func (s *GeneratorSuite) TestGeneratorHavingNoNamesOnArguments() {
	expected := `// KeyManager is an autogenerated failing mock type for the KeyManager type
type KeyManager struct {
	err error
}

// NewKeyManager creates a new KeyManager type instance
func NewKeyManager(err error) *KeyManager {
	return &KeyManager{err: err}
}

// GetKey provides a failing mock function with given fields: _a0, _a1
func (_m *KeyManager) GetKey(_a0 string, _a1 uint16) ([]byte, *test.Err) {
	var r0 []byte
	var r1 *test.Err

	return r0, r1
}
`
	s.checkGeneration(
		filepath.Join(fixturePath, "custom_error.go"), "KeyManager", false,
		expected,
	)
}

func (s *GeneratorSuite) TestGeneratorElidedType() {
	expected := `// RequesterElided is an autogenerated failing mock type for the RequesterElided type
type RequesterElided struct {
	err error
}

// NewRequesterElided creates a new RequesterElided type instance
func NewRequesterElided(err error) *RequesterElided {
	return &RequesterElided{err: err}
}

// Get provides a failing mock function with given fields: path, url
func (_m *RequesterElided) Get(path string, url string) error {
	var r0 error
	r0 = _m.err

	return r0
}
`
	s.checkGeneration(
		filepath.Join(fixturePath, "requester_elided.go"), "RequesterElided",
		false, expected,
	)
}

func (s *GeneratorSuite) TestGeneratorReturnElidedType() {
	expected := `// RequesterReturnElided is an autogenerated failing mock type for the RequesterReturnElided type
type RequesterReturnElided struct {
	err error
}

// NewRequesterReturnElided creates a new RequesterReturnElided type instance
func NewRequesterReturnElided(err error) *RequesterReturnElided {
	return &RequesterReturnElided{err: err}
}

// Get provides a failing mock function with given fields: path
func (_m *RequesterReturnElided) Get(path string) (int, int, int, error) {
	var r0 int
	var r1 int
	var r2 int
	var r3 error
	r3 = _m.err

	return r0, r1, r2, r3
}
`
	s.checkGeneration(
		filepath.Join(fixturePath, "requester_ret_elided.go"),
		"RequesterReturnElided", false, expected,
	)
}

func (s *GeneratorSuite) TestGeneratorVariadicArgs() {
	// Read the expected output from a "golden" file that we can also import in CompatSuite
	// to asserts its actual behavior.
	//
	// To regenerate the golden file: make fixture
	expectedBytes, err := ioutil.ReadFile(filepath.Join(fixturePath, "mocks", "requester_variadic.go"))
	s.NoError(err)
	expected := string(expectedBytes)
	expected = expected[strings.Index(expected, "// RequesterVariadic is"):]
	s.checkGeneration(
		filepath.Join(fixturePath, "requester_variadic.go"),
		"RequesterVariadic", false, expected,
	)
}

func (s *GeneratorSuite) TestGeneratorFuncType() {
	expected := `// Fooer is an autogenerated failing mock type for the Fooer type
type Fooer struct {
	err error
}

// NewFooer creates a new Fooer type instance
func NewFooer(err error) *Fooer {
	return &Fooer{err: err}
}

// Bar provides a failing mock function with given fields: f
func (_m *Fooer) Bar(f func([]int)) {
}

// Baz provides a failing mock function with given fields: path
func (_m *Fooer) Baz(path string) func(string) string {
	var r0 func(string) string

	return r0
}

// Foo provides a failing mock function with given fields: f
func (_m *Fooer) Foo(f func(string) string) error {
	var r0 error
	r0 = _m.err

	return r0
}
`
	s.checkGeneration(
		filepath.Join(fixturePath, "func_type.go"), "Fooer", false, expected,
	)
}

func (s *GeneratorSuite) TestGeneratorChanType() {
	expected := `// AsyncProducer is an autogenerated failing mock type for the AsyncProducer type
type AsyncProducer struct {
	err error
}

// NewAsyncProducer creates a new AsyncProducer type instance
func NewAsyncProducer(err error) *AsyncProducer {
	return &AsyncProducer{err: err}
}

// Input provides a failing mock function with given fields:
func (_m *AsyncProducer) Input() chan<- bool {
	var r0 chan<- bool

	return r0
}

// Output provides a failing mock function with given fields:
func (_m *AsyncProducer) Output() <-chan bool {
	var r0 <-chan bool

	return r0
}

// Whatever provides a failing mock function with given fields:
func (_m *AsyncProducer) Whatever() chan bool {
	var r0 chan bool

	return r0
}
`
	s.checkGeneration(
		filepath.Join(fixturePath, "async.go"), "AsyncProducer", false,
		expected,
	)
}

func (s *GeneratorSuite) TestGeneratorFromImport() {
	expected := `// MyReader is an autogenerated failing mock type for the MyReader type
type MyReader struct {
	err error
}

// NewMyReader creates a new MyReader type instance
func NewMyReader(err error) *MyReader {
	return &MyReader{err: err}
}

// Read provides a failing mock function with given fields: p
func (_m *MyReader) Read(p []byte) (int, error) {
	var r0 int
	var r1 error
	r1 = _m.err

	return r0, r1
}
`
	s.checkGeneration(
		filepath.Join(fixturePath, "io_import.go"), "MyReader", false,
		expected,
	)
}

func (s *GeneratorSuite) TestGeneratorComplexChanFromConsul() {
	expected := `// ConsulLock is an autogenerated failing mock type for the ConsulLock type
type ConsulLock struct {
	err error
}

// NewConsulLock creates a new ConsulLock type instance
func NewConsulLock(err error) *ConsulLock {
	return &ConsulLock{err: err}
}

// Lock provides a failing mock function with given fields: _a0
func (_m *ConsulLock) Lock(_a0 <-chan struct{}) (<-chan struct{}, error) {
	var r0 <-chan struct{}
	var r1 error
	r1 = _m.err

	return r0, r1
}

// Unlock provides a failing mock function with given fields:
func (_m *ConsulLock) Unlock() error {
	var r0 error
	r0 = _m.err

	return r0
}
`
	s.checkGeneration(
		filepath.Join(fixturePath, "consul.go"), "ConsulLock", false, expected,
	)
}

func (s *GeneratorSuite) TestGeneratorForEmptyInterface() {
	expected := `// Blank is an autogenerated failing mock type for the Blank type
type Blank struct {
	err error
}

// NewBlank creates a new Blank type instance
func NewBlank(err error) *Blank {
	return &Blank{err: err}
}

// Create provides a failing mock function with given fields: x
func (_m *Blank) Create(x interface{}) error {
	var r0 error
	r0 = _m.err

	return r0
}
`
	s.checkGeneration(
		filepath.Join(fixturePath, "empty_interface.go"), "Blank", false,
		expected,
	)
}

func (s *GeneratorSuite) TestGeneratorForMapFunc() {
	expected := `// MapFunc is an autogenerated failing mock type for the MapFunc type
type MapFunc struct {
	err error
}

// NewMapFunc creates a new MapFunc type instance
func NewMapFunc(err error) *MapFunc {
	return &MapFunc{err: err}
}

// Get provides a failing mock function with given fields: m
func (_m *MapFunc) Get(m map[string]func(string) string) error {
	var r0 error
	r0 = _m.err

	return r0
}
`
	s.checkGeneration(
		filepath.Join(fixturePath, "map_func.go"), "MapFunc", false, expected,
	)
}

func (s *GeneratorSuite) TestGeneratorForMethodUsingInterface() {
	expected := `// UsesOtherPkgIface is an autogenerated failing mock type for the UsesOtherPkgIface type
type UsesOtherPkgIface struct {
	err error
}

// NewUsesOtherPkgIface creates a new UsesOtherPkgIface type instance
func NewUsesOtherPkgIface(err error) *UsesOtherPkgIface {
	return &UsesOtherPkgIface{err: err}
}

// DoSomethingElse provides a failing mock function with given fields: obj
func (_m *UsesOtherPkgIface) DoSomethingElse(obj test.Sibling) {
}
`
	s.checkGeneration(
		filepath.Join(fixturePath, "mock_method_uses_pkg_iface.go"),
		"UsesOtherPkgIface", false, expected,
	)
}

func (s *GeneratorSuite) TestGeneratorForMethodUsingInterfaceInPackage() {
	expected := `// FailingUsesOtherPkgIface is an autogenerated failing mock type for the UsesOtherPkgIface type
type FailingUsesOtherPkgIface struct {
	err error
}

// NewFailingUsesOtherPkgIface creates a new FailingUsesOtherPkgIface type instance
func NewFailingUsesOtherPkgIface(err error) *FailingUsesOtherPkgIface {
	return &FailingUsesOtherPkgIface{err: err}
}

// DoSomethingElse provides a failing mock function with given fields: obj
func (_m *FailingUsesOtherPkgIface) DoSomethingElse(obj Sibling) {
}
`
	s.checkGeneration(
		filepath.Join(fixturePath, "mock_method_uses_pkg_iface.go"),
		"UsesOtherPkgIface", true, expected,
	)
}

func (s *GeneratorSuite) TestGeneratorWithAliasing() {
	expected := `// Example is an autogenerated failing mock type for the Example type
type Example struct {
	err error
}

// NewExample creates a new Example type instance
func NewExample(err error) *Example {
	return &Example{err: err}
}

// A provides a failing mock function with given fields:
func (_m *Example) A() http.Flusher {
	var r0 http.Flusher

	return r0
}

// B provides a failing mock function with given fields: _a0
func (_m *Example) B(_a0 string) fixtureshttp.MyStruct {
	var r0 fixtureshttp.MyStruct

	return r0
}
`
	s.checkGeneration(
		filepath.Join(fixturePath, "same_name_imports.go"), "Example", false,
		expected,
	)
}

func (s *GeneratorSuite) TestGeneratorWithImportSameAsLocalPackageInpkgNoCycle() {
	iface := s.getInterfaceFromFile("imports_same_as_package.go", "ImportsSameAsPackage")
	pkg := iface.QualifiedName
	gen := NewGenerator(iface, pkg, true)
	gen.GeneratePrologue(pkg)
	s.NotContains(gen.buf.String(), `import test "github.com/kyma-project/kyma/tools/failery/failery/fixtures/test"`)
}

func (s *GeneratorSuite) TestGeneratorWithImportSameAsLocalPackage() {
	expected := `// ImportsSameAsPackage is an autogenerated failing mock type for the ImportsSameAsPackage type
type ImportsSameAsPackage struct {
	err error
}

// NewImportsSameAsPackage creates a new ImportsSameAsPackage type instance
func NewImportsSameAsPackage(err error) *ImportsSameAsPackage {
	return &ImportsSameAsPackage{err: err}
}

// A provides a failing mock function with given fields:
func (_m *ImportsSameAsPackage) A() test.B {
	var r0 test.B

	return r0
}

// B provides a failing mock function with given fields:
func (_m *ImportsSameAsPackage) B() fixtures.KeyManager {
	var r0 fixtures.KeyManager

	return r0
}

// C provides a failing mock function with given fields: _a0
func (_m *ImportsSameAsPackage) C(_a0 fixtures.C) {
}
`
	s.checkGeneration(
		"imports_same_as_package.go", "ImportsSameAsPackage", false,
		expected,
	)
}

func (s *GeneratorSuite) TestPrologueWithImportSameAsLocalPackage() {
	generator := s.getGenerator(
		"imports_same_as_package.go", "ImportsSameAsPackage", false,
	)
	expected := `package mocks

import fixtures "` + generator.iface.QualifiedName + `"
import test "github.com/kyma-project/kyma/tools/failery/failery/fixtures/test"

`
	s.checkPrologueGeneration(generator, expected)
}

func (s *GeneratorSuite) TestPrologueWithImportFromNestedInterface() {
	generator := s.getGenerator(
		"imports_from_nested_interface.go", "HasConflictingNestedImports", false,
	)
	expected := `package mocks

import fixtureshttp "github.com/kyma-project/kyma/tools/failery/failery/fixtures/http"
import http "net/http"
import test "github.com/kyma-project/kyma/tools/failery/failery/fixtures"

`

	s.checkPrologueGeneration(generator, expected)
}

func (s *GeneratorSuite) TestGeneratorForStructValueReturn() {
	expected := `// A is an autogenerated failing mock type for the A type
type A struct {
	err error
}

// NewA creates a new A type instance
func NewA(err error) *A {
	return &A{err: err}
}

// Call provides a failing mock function with given fields:
func (_m *A) Call() (test.B, error) {
	var r0 test.B
	var r1 error
	r1 = _m.err

	return r0, r1
}
`
	s.checkGeneration(
		filepath.Join(fixturePath, "struct_value.go"), "A", false,
		expected,
	)
}

func TestGeneratorSuite(t *testing.T) {
	generatorSuite := new(GeneratorSuite)
	suite.Run(t, generatorSuite)
}
